unit uPreprocess;

interface

uses
  Classes, SysUtils, mwGenericLex, uGenLexer;

type
  EPreprocessorException = class(Exception)
  private
    fLineNo : integer;
  public
    constructor Create(const msg : string; const lineno : integer);
    property LineNo : integer read fLineNo;
  end;

  { TMapList }

  TMapList = class(TStringList)
  private
    fConsiderCase: boolean;
    function GetMapValue(index: integer): string;
    procedure SetConsiderCase(const AValue: boolean);
    procedure SetMapValue(index: integer; const Value: string);
  protected
{$IFDEF FPC}
    function DoCompareText(const s1,s2 : string) : PtrInt; override;
{$ENDIF}
  public
    constructor Create;
    function  AddEntry(const aName, aValue : string) : integer;
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    property MapValue[index : integer] : string read GetMapValue write SetMapValue;
    property ConsiderCase : boolean read fConsiderCase write SetConsiderCase;
  end;

  TLangPreprocessor = class
  private
    IncludeDirs : TStringList;
    MacroDefs : TMapList;
    MacroFuncArgs : TMapList;
    fLevel : integer;
    fLevelIgnore : TList;
    fGLC : TGenLexerClass;
{
    fVarI : integer;
    fVarJ : integer;
    procedure SetThreadName(const name : string);
    procedure SetCurrentFile(const name : string);
    procedure SetVarJ(val : integer);
    procedure SetVarI(val : integer);
    procedure LoadStandardMacros;
}
    procedure DoProcessStream(const name: string; lineNo: integer;
      Stream: TMemoryStream; OutStrings: TStrings);
    function GetLevelIgnoreValue(const lineno : integer): boolean;
    procedure SwitchLevelIgnoreValue(const lineno : integer);
    function ProcessIdentifier(Lex: TGenLexer; lineNo: integer): string;
    function ReplaceTokens(const tokenstring: string; lineNo : integer;
      bArgs : boolean): string;
    function ProcessMacroDefinition(bUndefine : boolean; Lex: TGenLexer;
      var lineNo: integer): integer;
    function GenLexerType : TGenLexerClass;
  public
    class procedure PreprocessStrings(GLType : TGenLexerClass; const fname : string; aStrings : TStrings);
    class procedure PreprocessFile(GLType : TGenLexerClass; const fin, fout : string);
    constructor Create(GLType : TGenLexerClass; const defIncDir : string);
    destructor Destroy; override;
    procedure Preprocess(const fname: string; aStrings: TStrings);
    procedure AddIncludeDirs(aStrings : TStrings);
  end;

implementation

uses
  uVersionInfo;

function CountLineEndings(ablock : string) : integer;
var
  tmpSL : TStringList;
begin
  tmpSL := TStringList.Create;
  try
    tmpSL.Text := ablock+'_';
    Result := tmpSL.Count - 1;
  finally
    tmpSL.Free;
  end;
end;

{ TLangPreprocessor }

class procedure TLangPreprocessor.PreprocessStrings(GLType : TGenLexerClass;
  const fname : string; aStrings : TStrings);
var
  P : TLangPreprocessor;
begin
  P := TLangPreprocessor.Create(GLType, ExtractFilePath(fname));
  try
    P.Preprocess(fname, aStrings);
  finally
    P.Free;
  end;
end;

class procedure TLangPreprocessor.PreprocessFile(GLType : TGenLexerClass;
  const fin, fout : string);
var
  SL : TStringList;
begin
  SL := TStringList.Create;
  try
    SL.LoadFromFile(fin);
    TLangPreprocessor.PreprocessStrings(GLType, fin, SL);
    SL.SaveToFile(fout);
  finally
    SL.Free;
  end;
end;

procedure TLangPreprocessor.Preprocess(const fname : string; aStrings : TStrings);
var
  Stream: TMemoryStream;
begin
//  MacroDefs.Clear;
  fLevelIgnore.Clear;

  Stream := TMemoryStream.Create;
  try
    aStrings.SaveToStream(Stream);
    aStrings.Clear;
    IncludeDirs.Add(ExtractFilePath(fname));
    fLevelIgnore.Add(nil); // level zero is NOT ignored (ignore is false == nil)
    fLevel := 0; // starting level is zero
    DoProcessStream(fname, 0, Stream, aStrings);
  finally
    Stream.Free;
  end;
end;

resourcestring
  sInvalidPreprocDirective = 'Invalid preprocessor directive';
  sIncludeNotFound = 'Unable to find include file: "%s"';
  sIncludeMissingQuotes = '#include directive requires a filename (e.g., #include "foo.h")';
  sMacroMismatch = 'Preprocessor macro function does not match instance';
  sUnmatchedDirective = 'Unmatched preprocessor directive';

function TLangPreprocessor.GetLevelIgnoreValue(const lineno : integer) : boolean;
begin
  if (fLevel < 0) or (fLevel >= fLevelIgnore.Count) then
    raise EPreprocessorException.Create(sInvalidPreprocDirective, lineno);
  Result := Assigned(fLevelIgnore[fLevel]);
end;

procedure TLangPreprocessor.SwitchLevelIgnoreValue(const lineno : integer);
begin
  if (fLevel < 0) or (fLevel >= fLevelIgnore.Count) then
    raise EPreprocessorException.Create(sInvalidPreprocDirective, lineno);
  if GetLevelIgnoreValue(lineno) then
    fLevelIgnore[fLevel] := nil
  else
    fLevelIgnore[fLevel] := Self;
end;

procedure AddOneOrMoreLines(OutStrings : TStrings; const S, name : string; var lineNo : integer);
var
  SL : TStringList;
begin
  if Pos(#13#10, S) > 0 then
  begin
    SL := TStringList.Create;
    try
      SL.Text := S;
      OutStrings.AddStrings(SL);
      // at the end of each multi-line macro expansion output a #line directive
      OutStrings.Add('#line ' + IntToStr(lineNo{ - SL.Count + 1}) + ' "' + name + '"');
//      lineNo := lineNo + SL.Count;
    finally
      SL.Free;
    end;
  end
  else
    OutStrings.Add(S);
end;

function TLangPreprocessor.ReplaceTokens(const tokenstring: string;
  lineNo : integer; bArgs : boolean): string;
var
  Lex : TGenLexer;
  map : TMapList;
   
  procedure AddToResult;
  var
    i : integer;
  begin
    if Lex.Id = piIdent then
    begin
      if bArgs then
      begin
        i := map.IndexOf(Lex.Token);
        if i <> -1 then
          Result := Result + map.MapValue[i]
        else
          Result := Result + Lex.Token;
      end
      else
      begin
        // pass lex to ProcessIdentifier
        Result := Result + ProcessIdentifier(Lex, lineNo);
      end;
    end
    else
      Result := Result + Lex.Token;
  end;
begin
  Result := '';
  if bArgs then
    map := MacroFuncArgs
  else
    map := MacroDefs;
  Lex := GenLexerType.CreateLexer;
  try
    Lex.SetStartData(@tokenstring[1], Length(tokenstring));
    while not Lex.AtEnd do
    begin
      AddToResult;
      Lex.Next;
    end;
    if Lex.Id <> piUnknown then
      AddToResult;
  finally
    Lex.Free;
  end;
end;

function TLangPreprocessor.ProcessIdentifier(Lex : TGenLexer; lineNo : integer) : string;
var
  i : integer;
  macroVal, dirText, prevToken, tmp, tok : string;
begin
  Result := '';
  // is token in defmap?  if so replace it (smartly)
  // token may be a function which takes parameters
  // so if it is then we need to grab more tokens from ( to )
  tok := Lex.Token;
(*
  if tok = '__LINE__' then
  begin
    Result := IntToStr(lineNo);
  end
  else if tok = '__ResetI__' then
  begin
    SetVarI(0);
    Result := '';
  end
  else if tok = '__ResetJ__' then
  begin
    SetVarJ(0);
    Result := '';
  end
  else if tok = '__IncI__' then
  begin
    SetVarI(fVarI+1);
    Result := '';
  end
  else if tok = '__IncJ__' then
  begin
    SetVarJ(fVarJ+1);
    Result := '';
  end
  else if tok = '__DecI__' then
  begin
    SetVarI(fVarI-1);
    Result := '';
  end
  else if tok = '__DecJ__' then
  begin
    SetVarJ(fVarJ-1);
    Result := '';
  end
  else
  begin
*)
    i := MacroDefs.IndexOf(tok);
    if i <> -1 then
    begin
      macroVal := MacroDefs.MapValue[i];
      if Pos('#', macroVal) = 1 then
      begin
        // function macro - complicated
        // format of macroVal == #arg1,arg2,...,argn#formula
        // is next token '('?
        Lex.Next;
        if (Lex.Id = piSymbol) and (Lex.Token = '(') then
        begin
          // move past the '('
          Lex.Next;
          MacroFuncArgs.Clear;  // start with an empty arg map
          Delete(macroVal, 1, 1); // remove starting '#'
          i := Pos('#', macroVal);
          dirText := Copy(macroVal, 1, i-1)+ ','; // arg1,arg2,...,argn,
          macroVal := Copy(macroVal, i+1, MaxInt);
          while (Lex.Id <> piLineEnd) and not Lex.AtEnd do
          begin
            // skip whitespace prior to each arg instance
            while (Lex.Id = piSpace) and not Lex.AtEnd do
              Lex.Next;
            prevToken := '';
            // now collect tokens until either ',' or ')' or whitespace
            while not ((Lex.Id = piSpace) or
                       ((Lex.Id = piSymbol) and
                        ((Lex.Token = ',') or (Lex.Token = ')')))) do
            begin
              prevToken := prevToken + Lex.Token;
              Lex.Next;
            end;
            // now match instance to arg
            // each identifier in the instance maps to
            // an argument in the defined macro function
            i := Pos(',', dirText);
            if i = 0 then Break;
            MacroFuncArgs.AddEntry(Copy(dirText, 1, i-1), prevToken);
            Delete(dirText, 1, i);
            // skip whitespace following each arg instance
            while (Lex.Id = piSpace) and not Lex.AtEnd do
              Lex.Next;
            if (Lex.Id = piSymbol) and (Lex.Token = ')') then
              Break; // stop looping
            Lex.Next;
          end;
          if (i = 0) or (dirText <> '') or (Lex.Id = piLineEnd) then
            raise EPreprocessorException.Create(sMacroMismatch, lineNo);
          // we have eaten '(' through ')' (possibly to end of line)
          tmp := ReplaceTokens(macroVal, lineNo, True);
          Result := ReplaceTokens(tmp, lineNo, False);
        end
        else
          raise EPreprocessorException.Create(sMacroMismatch, lineNo);
      end
      else
      begin
        // simple macro substitution
        Result := ReplaceTokens(macroVal, lineNo, False);
  //      Result := macroVal;
      end;
    end
    else
      Result := tok;
//  end;
end;

function TLangPreprocessor.ProcessMacroDefinition(bUndefine : boolean;
  Lex : TGenLexer; var lineNo : integer) : integer;
var
  macro, dirText, macroVal, prevToken : string;
  bEndOfDefine, bArgMode : boolean;
  i, prevId : integer;

  procedure HandleDefinition;
  begin
    macroVal := '';
    bEndOfDefine := (Lex.Id = piLineEnd) or Lex.AtEnd;
    while not bEndOfDefine do
    begin
      prevId := Lex.Id;
      prevToken := Lex.Token;
      Lex.Next;
      if bArgMode then
      begin
        if prevId = piIdent then
        begin
          dirText := dirText + prevToken + ','; // add argument
        end
        else if (prevId = piSymbol) and (prevToken = ')') then
        begin
          Delete(dirText, Length(dirText), 1); // remove last character
          dirText := dirText + '#'; // end arguments
          bArgMode := False;
        end;
      end
      else
      begin
        if prevId <> piComment then
        begin
          if (prevId = piSymbol) and (prevToken = '\') and (Lex.Id = piLineEnd) then
          begin
            inc(lineNo);
            inc(Result);
            continue;
          end
          else
            macroVal := macroVal + prevToken;
        end;
      end;
      bEndOfDefine := (Lex.Id = piLineEnd) or Lex.AtEnd;
    end;
  end;
begin
  Result := 0;
  // add/remove definition to/from defmap
  // line should have a space token (1) and then an identifier
  // token followed by an optional parameter list ( ... )
  // followed by an optional value (everything left on the line
  // to EOL -- or multiple lines if line ends with '\')
  Lex.Next; // token ID should be piSpace
  if Lex.Id <> piSpace then
    raise EPreprocessorException.Create(sInvalidPreprocDirective, lineNo);
  Lex.Next; // token ID should be piIdent
  if Lex.Id <> piIdent then
    raise EPreprocessorException.Create(sInvalidPreprocDirective, lineNo);
  macro := Lex.Token; // the macro name
  if bUndefine then
  begin
    // no more info needed.  Remove macro
    i := MacroDefs.IndexOf(macro);
    if i <> -1 then
      MacroDefs.Delete(i);
  end
  else
  begin
    // is the next token whitespace?
    Lex.Next;
    if Lex.Id in [piSpace, piLineEnd] then
    begin
      // collect to end of #define (skipping comments)
      bArgMode := False;
      dirText := '';
      HandleDefinition;
    end
    else if (Lex.Id = piSymbol) and (Lex.Token = '(') then
    begin
      bArgMode := True;
      Lex.Next;
      dirText := '#'; // flag that this is a function macro
      HandleDefinition;
    end;
    dirText := Trim(dirText) + TrimLeft(macroVal);
    MacroDefs.AddEntry(macro, dirText);
  end;
end;

procedure TLangPreprocessor.DoProcessStream(const name : string; lineNo : integer;
  Stream : TMemoryStream; OutStrings : TStrings);
var
  Lex: TGenLexer;
  S, dir, dirText, tmpname, usePath, macro : string;
  X : TMemoryStream;
  i, cnt, origLevel : integer;
  bFileFound, bDefined, bProcess{, bEatToEOL, bSetThreadName} : boolean;
begin
  S := '';
//  SetCurrentFile(name);
  origLevel := fLevel;
//  bSetThreadName := False;
  // at the start of each call to this function output a #line directive
  bProcess := not GetLevelIgnoreValue(lineNo);
  if bProcess and (lineNo > 0) then
    OutStrings.Add('#line ' + IntToStr(lineNo) + ' "' + name + '"');
  if lineNo <= 0 then
    lineNo := 1;
  Lex := GenLexerType.CreateLexer;
  try
    Lex.SetStartData(Stream.Memory, Integer(Stream.Size));
    while not Lex.AtEnd do
    begin
      case Lex.Id of
        piSpace : begin
          S := S + ' '; // all space == single space
        end;
        piLineEnd, piInnerLineEnd : begin
          // output S and clear it
          if bProcess then
            AddOneOrMoreLines(OutStrings, S, name, lineNo)
          else
            OutStrings.Add('');
          S := '';
          inc(lineNo);
        end;
        piComment : begin
          // strip all comments
        end;
        piDirective : begin
          // some sort of preprocessor directive.
//          bEatToEOL := True;
          dir := AnsiLowerCase(Lex.Token);
          dirText := '';
          if dir = '#include' then
          begin
            // collect to end of line
            Lex.Next;
            while (Lex.Id <> piLineEnd) and not Lex.AtEnd do
            begin
              dirText := dirText + Lex.Token;
              Lex.Next;
            end;
            if bProcess then
            begin
              dirText := Trim(dirText);
              // get filename
              if Pos('"', dirText) > 0 then
              begin
                tmpName := StringReplace(dirText, '"', '', [rfReplaceAll]);
                X := TMemoryStream.Create;
                try
                  usePath := '';
                  bFileFound := False;
                  for i := 0 to IncludeDirs.Count - 1 do
                  begin
                    usePath := IncludeTrailingPathDelimiter(IncludeDirs[i]);
                    bFileFound := FileExists(usePath+tmpName);
                    if bFileFound then Break;
                  end;
                  if bFileFound then
                  begin
                    // load into stream
                    X.LoadFromFile(usePath+tmpName);
                    // call function recursively
                    DoProcessStream(tmpName, 1, X, OutStrings);
                  end
                  else
                    raise EPreprocessorException.Create(Format(sIncludeNotFound, [tmpName]), lineNo);
                finally
                  X.Free;
                end;
              end
              else
                raise EPreprocessorException.Create(sIncludeMissingQuotes, lineNo);
              // at the end of each #include output a #line directive
              OutStrings.Add('#line ' + IntToStr(lineNo) + ' "' + name + '"');
//              SetCurrentFile(name);
            end;
          end
          else if (dir = '#define') or (dir = '#undef') then
          begin
            cnt := 1; // number of lines in #define
            if bProcess then
              cnt := cnt + ProcessMacroDefinition(dir = '#undef', Lex, lineNo);
            // output blank line(s) to replace #define
            for i := 0 to cnt - 1 do
              OutStrings.Add('');
          end
          else if (dir = '#ifndef') or (dir = '#ifdef') then
          begin
            // increment level
            Lex.Next; // token ID should be piSpace
            if Lex.Id <> piSpace then
              raise EPreprocessorException.Create(sInvalidPreprocDirective, lineNo);
            Lex.Next; // token ID should be piIdent or piNumber
            if not Lex.Id in [piIdent, piNumber] then
              raise EPreprocessorException.Create(sInvalidPreprocDirective, lineNo);
            if Lex.Id = piIdent then
            begin
              macro := Lex.Token; // the macro name
              bDefined := MacroDefs.IndexOf(macro) <> -1;
            end
            else
            begin
              // Lex.Token == number
              i := StrToIntDef(Lex.Token, 0);
              bDefined := i <> 0;
            end;
            if (bDefined and (dir = '#ifdef')) or not (bDefined or (dir = '#ifdef')) then
              fLevelIgnore.Add(nil)
            else
              fLevelIgnore.Add(Self);
            inc(fLevel);
            bProcess := not GetLevelIgnoreValue(lineNo);
            // output a blank line to replace directive
            OutStrings.Add('');
          end
          else if dir = '#endif' then
          begin
            // decrement level
            fLevelIgnore.Delete(fLevel);
            dec(fLevel);
            bProcess := not GetLevelIgnoreValue(lineNo);
            // output a blank line to replace directive
            OutStrings.Add('');
          end
{
          else if dir = '##' then
          begin
            // this token is simply deleted and we do not delete to end of line
            bEatToEOL := False;
          end
}
          else if dir = '#else' then
          begin
            // switch mode at current level (must be > 0)
            SwitchLevelIgnoreValue(lineNo);
            bProcess := not bProcess;
            // output a blank line to replace directive
            OutStrings.Add('');
          end;
          // ignore #pragma and #line directives
          // eat to end of line
//          if bEatToEOL then
//          begin
            while (Lex.Id <> piLineEnd) and not Lex.AtEnd do
              Lex.Next;
            inc(lineNo);
//          end;
        end;
        piZero, piUnknown : {do nothing};
{
        piKeyword : begin
          if bProcess then
          begin
            dir := Lex.Token;
            bSetThreadName := (dir = 'subroutine') or (dir = 'thread');
            S := S + Lex.Token;
          end;
        end;
}
        piIdent : begin
          if bProcess then
            S := S + ProcessIdentifier(Lex, lineNo);
{
          begin
            dir := ProcessIdentifier(Lex, lineNo);
            if bSetThreadName then
            begin
              SetThreadName(dir);
              bSetThreadName := False;
            end;
            S := S + dir;
          end;
}
        end;
      else
        if bProcess then
          S := S + Lex.Token;
      end;
      Lex.Next;
    end;
    if fLevel <> origLevel then
      raise EPreprocessorException.Create(sUnmatchedDirective, lineNo);
    if (S <> '') and bProcess then
      AddOneOrMoreLines(OutStrings, S, name, lineNo);
  finally
    Lex.Free;
  end;
end;

constructor TLangPreprocessor.Create(GLType : TGenLexerClass; const defIncDir : string);
begin
  inherited Create;
  fGLC := GLType;
  IncludeDirs := TStringList.Create;
  IncludeDirs.Add(defIncDir);
  MacroDefs := TMapList.Create;
  MacroFuncArgs := TMapList.Create;
  fLevelIgnore := TList.Create;
//  fVarI := 0;
//  fVarJ := 0;
//  LoadStandardMacros;
end;

destructor TLangPreprocessor.Destroy;
begin
  IncludeDirs.Clear;
  MacroDefs.Clear;
  MacroFuncArgs.Clear;
  FreeAndNil(IncludeDirs);
  FreeAndNil(MacroDefs);
  FreeAndNil(MacroFuncArgs);
  FreeAndNil(fLevelIgnore);
  inherited;
end;

procedure TLangPreprocessor.AddIncludeDirs(aStrings: TStrings);
begin
  IncludeDirs.AddStrings(aStrings);
end;

function TLangPreprocessor.GenLexerType: TGenLexerClass;
begin
  Result := fGLC;
end;

(*
procedure TLangPreprocessor.SetThreadName(const name: string);
var
  idx : integer;
begin
  idx := MacroDefs.IndexOf('__THREADNAME__');
  if idx <> -1 then
    MacroDefs.MapValue[idx] := name;
  SetVarJ(0);
  SetVarI(0);
end;

procedure TLangPreprocessor.LoadStandardMacros;
begin
{
  MacroDefs.AddEntry('__ResetI__', '');
  MacroDefs.AddEntry('__ResetJ__', '');
  MacroDefs.AddEntry('__IncI__', '');
  MacroDefs.AddEntry('__IncJ__', '');
  MacroDefs.AddEntry('__DecI__', '');
  MacroDefs.AddEntry('__DecJ__', '');
  MacroDefs.AddEntry('__LINE__', '0');
}
  MacroDefs.AddEntry('__I__', '0');
  MacroDefs.AddEntry('__J__', '0');
  MacroDefs.AddEntry('__THREADNAME__', '');
  MacroDefs.AddEntry('__FILE__', '''''');
  MacroDefs.AddEntry('__VER__', ''''+GetProductVersion+'''');
end;

procedure TLangPreprocessor.SetCurrentFile(const name: string);
var
  idx : integer;
begin
  idx := MacroDefs.IndexOf('__FILE__');
  if idx <> -1 then
    MacroDefs.MapValue[idx] := ''''+name+'''';
end;

procedure TLangPreprocessor.SetVarJ(val: integer);
var
  idx : integer;
begin
  fVarJ := val;
  idx := MacroDefs.IndexOf('__J__');
  if idx <> -1 then
    MacroDefs.MapValue[idx] := IntToStr(fVarJ);
end;

procedure TLangPreprocessor.SetVarI(val: integer);
var
  idx : integer;
begin
  fVarI := val;
  idx := MacroDefs.IndexOf('__I__');
  if idx <> -1 then
    MacroDefs.MapValue[idx] := IntToStr(fVarI);
end;
*)

{ EPreprocessorException }

constructor EPreprocessorException.Create(const msg: string;
  const lineno: integer);
begin
  inherited Create(msg);
  fLineNo := lineno;
end;

{ TMapList }

type
  TStrObj = class(TObject)
  public
    Value : string;
  end;

function TMapList.AddEntry(const aName, aValue: string): integer;
var
  obj : TStrObj;
begin
  Result := IndexOf(aName);
  if Result = -1 then
  begin
    obj := TStrObj.Create;
    Result := AddObject(aName, obj);
    obj.Value := aValue;
  end;
end;

procedure TMapList.Clear;
var
  i : integer;
begin
  for i := 0 to Count - 1 do
    Objects[i].Free;
  inherited;
end;

constructor TMapList.Create;
begin
  inherited;
  CaseSensitive := True;
  ConsiderCase  := True;
  Sorted := True;
  Duplicates := dupIgnore;
end;

procedure TMapList.Delete(Index: Integer);
begin
  Objects[Index].Free;
  Objects[Index] := nil;
  inherited;
end;

function TMapList.GetMapValue(index: integer): string;
begin
  Result := TStrObj(Objects[index]).Value;
end;

procedure TMapList.SetMapValue(index: integer; const Value: string);
begin
  TStrObj(Objects[index]).Value := Value;
end;

procedure TMapList.SetConsiderCase(const AValue: boolean);
begin
  if fConsiderCase=AValue then exit;
  fConsiderCase:=AValue;
  if Sorted then
    Sort;
end;

{$IFDEF FPC}
function TMapList.DoCompareText(const s1, s2: string): PtrInt;
begin
  if CaseSensitive or ConsiderCase then
    result := AnsiCompareStr(s1,s2)
  else
    result := AnsiCompareText(s1,s2);
end;
{$ENDIF}

end.
